from bhepop2.tools import (
    add_household_size_attribute,
    add_household_type_attribute,
)


def test_add_household_size_attribute(eqasim_population):
    pop = add_household_size_attribute(eqasim_population)
    assert set(pop["size"].unique()) == {"1_pers", "2_pers", "3_pers", "4_pers", "5_pers_or_more"}


def test_add_household_type_attribute(eqasim_population):
    pop = add_household_type_attribute(eqasim_population)
    assert set(pop["family_comp"].unique()) == {
        "Single_man",
        "Single_wom",
        "Couple_without_child",
        "Couple_with_child",
        "Single_parent",
        "complex_hh",
    }



from bhepop2.quantitative_enrichment import QuantitativeEnrichment

import numpy as np


def test_bhepop2_enrich(
    synthetic_population_nantes,
    filosofi_distributions_nantes,
    test_modalities,
    test_parameters,
    test_seed,
    expected_enriched_population_nantes,
):
    synthetic_population_nantes.drop("sex", axis=1, inplace=True)

    enrich_class = QuantitativeEnrichment(
        synthetic_population_nantes,
        filosofi_distributions_nantes,
        list(test_modalities.keys()),
        parameters=test_parameters,
        seed=test_seed,
    )

    enrich_class.optimise()

    pop = enrich_class.assign_feature_value_to_pop()

    # pop.to_csv("tests/data/nantes_enriched.csv", index=False)

    assert np.all((pop == expected_enriched_population_nantes).to_numpy())



from bhepop2.tools import read_filosofi, filosofi_attributes
from bhepop2.functions import get_attributes

import pytest
import pandas as pd
import os
import shutil

SEED = 42

CODE_INSEE = "44109"
MODALITIES = {
    "ownership": ["Owner", "Tenant"],
    "age": ["0_29", "30_39", "40_49", "50_59", "60_74", "75_or_more"],
    "size": ["1_pers", "2_pers", "3_pers", "4_pers", "5_pers_or_more"],
    "family_comp": [
        "Single_man",
        "Single_wom",
        "Couple_without_child",
        "Couple_with_child",
        "Single_parent",
        "complex_hh",
    ],
}

parameters = {
    "abs_minimum": 0,
    "relative_maximum": 1.5,
    "delta_min": 1000,
}

TEST_DATA_FOLDER = "tests/data/"
TMP_DIR = "tests/tmp/"


@pytest.fixture(scope="session")
def tmp_dir():
    return TMP_DIR


@pytest.fixture(scope="session", autouse=True)
def tmp_dir_create_delete(tmp_dir):
    if not os.path.exists(tmp_dir):
        os.mkdir(tmp_dir)

    yield

    shutil.rmtree(tmp_dir)


@pytest.fixture(scope="session")
def test_insee_code():
    return CODE_INSEE


@pytest.fixture(scope="session")
def test_modalities():
    return MODALITIES


@pytest.fixture(scope="session")
def test_attributes(test_modalities):
    return get_attributes(test_modalities)


@pytest.fixture(scope="session")
def test_parameters():
    return parameters


@pytest.fixture(scope="session")
def test_seed():
    return SEED


@pytest.fixture(scope="session")
def filosofi_distributions_nantes(test_insee_code):
    filosofi = read_filosofi(
        TEST_DATA_FOLDER + "FILO_DISP_COM.xls", "15", filosofi_attributes, [test_insee_code]
    )

    return filosofi


@pytest.fixture(scope="session")
def synthetic_population_nantes():
    return pd.read_csv(TEST_DATA_FOLDER + "synpop_nantes.csv")


@pytest.fixture(scope="session")
def pop_synt_men_nantes():
    synt_pop = pd.read_csv("tests/data/pop_synth_men_nantes.csv")
    return synt_pop


@pytest.fixture(scope="session")
def expected_enriched_population_nantes():
    return pd.read_csv(TEST_DATA_FOLDER + "nantes_enriched.csv")


@pytest.fixture(scope="session")
def eqasim_population():
    return pd.read_csv(TEST_DATA_FOLDER + "eqasim_population_0.001.csv", sep=";")


@pytest.fixture(scope="session")
def eqasim_households():
    return pd.read_csv(TEST_DATA_FOLDER + "eqasim_households_0.001.csv", sep=";")



import pandas as pd

from bhepop2.functions import *
import numpy as np

import pytest


def test_get_attributes(test_modalities):
    """
    Test that the returned attributes are correct.
    """

    assert get_attributes(test_modalities) == ["ownership", "age", "size", "family_comp"]


def test_modality_feature():
    """
    Test that the result is a function with expected behaviour.
    """

    attribute = "my_attribute"
    modality_0 = "modality_0"
    modality_1 = "modality_1"

    feature = modality_feature(attribute, modality_0)

    assert callable(feature)
    assert feature({attribute: modality_0})
    assert not feature({attribute: modality_1})


def test_compute_crossed_modalities_frequencies(
    synthetic_population_nantes, test_modalities, test_attributes
):
    """
    Test the crossed modalities frequencies have the correct columns and sum to 1.
    """
    freq_df = compute_crossed_modalities_frequencies(synthetic_population_nantes, test_modalities)

    assert list(freq_df.columns) == test_attributes + ["probability"]
    assert np.isclose(freq_df["probability"].sum(), 1)


def test_infer_modalities_from_distributions(filosofi_distributions_nantes):
    modalities = infer_modalities_from_distributions(filosofi_distributions_nantes)

    assert isinstance(modalities, dict)
    assert "all" not in modalities


@pytest.mark.parametrize(
    "delta_min, expected_length",
    [
        (None, 190),
        (1000, 41),
    ],
)
def test_compute_feature_values(
    delta_min, expected_length, filosofi_distributions_nantes, test_attributes
):
    """
    Test that the feature values list have the correct length and is sorted.
    """

    filo = filosofi_distributions_nantes[
        filosofi_distributions_nantes["attribute"].isin(test_attributes)
    ]
    feature_values = compute_feature_values(filo, 1.5, delta_min)

    assert len(feature_values) == expected_length
    assert feature_values == sorted(feature_values)


def test_get_feature_from_qualitative_distribution():
    distribution = pd.DataFrame(
        {
            "attribute": "age",
            "modality": ["10_17", "18_35"],
            "0voit": [0.99, 0.35],
            "1voit": [0.01, 0.5],
            "2voit": [0, 0.1],
            "3voit": [0, 0.05],
        }
    )

    features = get_feature_from_qualitative_distribution(distribution)

    assert features == ["0voit", "1voit", "2voit", "3voit"]

    distribution.loc[0, "2voit"] = 0.5

    with pytest.raises(AssertionError):
        get_feature_from_qualitative_distribution(distribution)


def test_interpolate_feature_prob():
    """
    Test that the feature probability interpolation is a float and has a coherent value.
    """

    distribution = [9794, 12961, 14914, 16865, 18687, 20763, 23357, 27069, 33514, 50271]
    feature_value = 16000
    interpolation = interpolate_feature_prob(feature_value, distribution)

    assert isinstance(interpolation, float)
    assert 0.3 <= interpolation <= 0.4



import pandas as pd
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from os import path

from bhepop2.functions import get_feature_from_qualitative_distribution

DEFAULT_PLOT_TITLE_FORMAT = "Modality {modality} from attribute {attribute}"
DEFAULT_OUTPUT_FOLDER = "outputs/"


class PopulationAnalysis:
    """
    Analysis class for synthetic populations.

    Synthetic populations must be identical except for their feature columns.

    The values of the feature columns and their distributions are compared between populations
    and to the reference distribution.

    Analysis is realised on the given modalities, which must be a subset of the modalities used for enrichment
    (and thus available in the population(s) and distributions).

    The following analysis are available:
        - Graphs comparing the distributions in the population(s) to the original distributions (one per modality)
        - A table describing the error of the population(s) in comparison to the distributions (one line per modality),
        ordered by number of individuals in the modality
    """

    # column describing the analysis class (depends on the feature's type)
    CLASS_COLUMN = None

    # column describing the value corresponding to the class
    VALUE_COLUMN = "value"

    def __init__(
        self,
        populations: dict,
        modalities: dict,
        feature_column: str,
        distributions: pd.DataFrame,
        distributions_name: str = "reference",
        plot_title_format: str = DEFAULT_PLOT_TITLE_FORMAT,
        output_folder: str = None,
    ):
        """
        Create an analysis instance containing the analysed populations and the reference distributions.

        :param populations: dict containing the analysed populations { pop_name: population_dataframe }
        :param modalities: attribute modalities dict { attribute: [modalities] }
        :param feature_column: name of the column containing the
        :param distributions:
        :param distributions_name:
        :param plot_title_format:
        :param output_folder:
        """
        # analysed data

        # check populations
        if len(populations) == 0:
            raise ValueError("No population to analyse")
        self.populations = populations

        # check distributions DataFrame
        self.distributions = distributions
        self.distributions_name = distributions_name

        # check modalities are consistent with distributions
        self.modalities = modalities

        # check feature column is in populations
        for name, population in populations.items():
            if feature_column not in population.columns:
                raise KeyError(
                    f"Feature column '{feature_column}' not found in '{name}' population"
                )
        self.feature_column = feature_column

        # output fields
        self._output_folder = None
        if output_folder:  # pragma: no cover
            self.set_output_folder(output_folder)

        self.plot_title_format = plot_title_format

        # analysis table
        self.analysis_table = self._evaluate_analysis_table()

    def set_output_folder(self, output_folder):
        """
        Set a new output folder for this analysis instance.

        :param output_folder: valid output folder path
        """
        if not path.isdir(output_folder):
            raise ValueError(f"{output_folder} is not a folder")
        self._output_folder = output_folder

    def assert_output_folder(self):
        """
        Check that the output folder is set.

        :raises: AssertionError
        """
        assert (
            self._output_folder is not None
        ), "No output folder is set, use set_output_folder method."

    def _evaluate_analysis_table(self):
        """
        Create a table used for comparing populations/distributions.

        The resulting DataFrame contains the following columns:
            - attribute: attribute name
            - modality: modality name
            - self.PROPORTION_COLUMN: value describing the proportion taken for the corresponding
            - one column with the observed_name value
            + one column for each population name

        :return: analysis DataFrame
        """

        # format distributions for analysis
        distributions_df = self._format_distributions_for_analysis()
        distributions_df["source"] = self.distributions_name

        # add population analysis dataframes
        analysis_list = [distributions_df]
        for name, population in self.populations.items():
            population_df = self._compute_distributions_by_attribute(population)
            population_df["source"] = name
            analysis_list.append(population_df)

        # create an analysis dataframe containing all deciles
        analysis_df = pd.concat(analysis_list)

        # pivot data to get final dataframe
        analysis_df = analysis_df.pivot(
            columns="source", index=["attribute", "modality", self.CLASS_COLUMN]
        ).reset_index()
        columns = list(analysis_df.columns.get_level_values(1))
        columns[0], columns[1], columns[2] = "attribute", "modality", self.CLASS_COLUMN
        analysis_df.columns = columns

        return analysis_df

    def _format_distributions_for_analysis(self):
        """
        Format the distributions table for as an analysis table.

        :return: distributions as an analysis table
        """
        raise NotImplementedError

    def _compute_distributions_by_attribute(self, population: pd.DataFrame) -> pd.DataFrame:
        """
        Compute the feature values distribution for each modality.

        Generate an analysis table for this population.

        :param population: population DataFrame

        :return: analysis table
        """

        # compute distribution for the whole population
        df_analysis = self._compute_distribution(population)
        df_analysis["attribute"] = "all"
        df_analysis["modality"] = "all"

        # distribution of each modality
        for attribute in self.modalities.keys():
            for modality in self.modalities[attribute]:
                distribution = self._compute_distribution(
                    population[population[attribute] == modality]
                )
                distribution["attribute"] = attribute
                distribution["modality"] = modality

                df_analysis = pd.concat([df_analysis, distribution])
        df_analysis["source"] = "enriched"
        return df_analysis

    def _compute_distribution(self, population: pd.DataFrame):
        """
        Get distribution of the feature values in the population.

        :param population: population DataFrame

        :return: analysis table of the population
        """
        raise NotImplementedError

    def generate_analysis_plots(self):
        """
        Generate plots comparing the population(s) to the original distributions (one per modality).

        Plots are exported to PNG images in the output folder.
        """

        self.assert_output_folder()

        # plot analysis for global distributions
        self.plot_analysis_compare("all", "all").write_image(
            path.join(self._output_folder, "all-all.png")
        )

        # plot analysis by attribute
        for attribute in self.modalities.keys():
            for modality in self.modalities[attribute]:
                self.plot_analysis_compare(attribute, modality).write_image(
                    path.join(self._output_folder, f"{attribute}-{modality}.png")
                )

    def plot_analysis_compare(self, attribute: str, modality: str):
        """
        Generate a plot comparing the populations and the distributions, for the given attribute and modality.

        :param attribute: attribute value
        :param modality: attribute modality

        :return: Plotly Figure
        """
        raise NotImplementedError

    def generate_analysis_error_table(self, export_csv: bool = True):
        """
        Generate a table describing how analysed populations deviate from the original distributions.

        :param export_csv:
        :return:
        """
        raise NotImplementedError

    def get_plot_title(self, attribute: str, modality: str) -> str:
        """
        Get the plot title for the given attribute and modality.

        Based on the plot_title_format parameter.

        :param attribute: attribute value
        :param modality: attribute modality

        :return: plot title
        """
        return self.plot_title_format.format(
            observed_name=self.distributions_name, attribute=attribute, modality=modality
        )


class QuantitativeAnalysis(PopulationAnalysis):
    CLASS_COLUMN = "decile"

    def plot_analysis_compare(self, attribute: str, modality: str):
        """
        Comparison plot between reference data and simulation

        :param attribute:
        :param modality:

        :return: Plotly figure
        """

        analysis_table = self.analysis_table
        analysis_table = analysis_table[
            (analysis_table["attribute"] == attribute) & (analysis_table["modality"] == modality)
        ]
        fig = px.line(
            x=analysis_table[self.distributions_name],
            y=analysis_table[self.distributions_name],
            color_discrete_sequence=["black"],
        )

        # add simulated names
        for population_name in self.populations.keys():
            fig.add_trace(
                go.Scatter(
                    x=analysis_table[self.distributions_name],
                    y=analysis_table[population_name],
                    mode="markers",
                    name=population_name,
                )
            )

        # configure plot
        fig.update_layout(
            title=self.get_plot_title(attribute, modality),
            xaxis_title=f"Observation ({self.distributions_name})",
            yaxis_title="Simulation",
        )

        return fig

    def generate_analysis_error_table(self, export_csv=True):
        """
        Generate a table describing how analysed populations deviate from the original distributions.

        :param export_csv: boolean a csv export should be realised

        :return: error table DataFrame
        """
        self.assert_output_folder()

        analysis_df = self.analysis_table
        # evaluate distance to distributions
        for population_name in self.populations.keys():
            analysis_df[f"{population_name}_perc_error"] = abs(
                (analysis_df[population_name] - analysis_df[self.distributions_name])
                / analysis_df[self.distributions_name]
            )
        error_analysis_df = analysis_df.groupby(["attribute", "modality"], as_index=False).agg(
            {f"{population_name}_perc_error": "mean" for population_name in self.populations.keys()}
        )

        # count number of occurrence of each modality and add it as a "number" column, sort descending
        pop = list(self.populations.values())[0]
        count = [
            pop.groupby([attribute])[attribute].count().rename("number")
            for attribute in self.modalities.keys()
        ]
        count = pd.concat(count)
        count["all"] = len(pop)
        error_analysis_df = error_analysis_df.merge(
            count, how="left", left_on="modality", right_index=True
        )
        error_analysis_df.sort_values(["number"], ascending=False, inplace=True)

        # export to csv if asked
        if export_csv:
            error_analysis_df.to_csv(
                f"{self._output_folder}/analysis_error.csv", sep=";", index=False
            )

        return error_analysis_df

    def _format_distributions_for_analysis(self):
        distributions = self.distributions[
            ["attribute", "modality", "D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8", "D9"]
        ]
        distributions_formated = distributions.melt(
            id_vars=["attribute", "modality"],
            value_vars=["D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8", "D9"],
            value_name=self.VALUE_COLUMN,
            var_name=self.CLASS_COLUMN,
        )
        return distributions_formated

    def _compute_distribution(self, population: pd.DataFrame) -> pd.DataFrame:
        """
        Compute decile distribution of the feature values.

        :param population: analysed population

        :return: dataframe of deciles
        """
        return pd.DataFrame(
            {
                self.VALUE_COLUMN: np.percentile(
                    population[self.feature_column],
                    np.arange(0, 100, 10),
                )[1:],
                self.CLASS_COLUMN: ["D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8", "D9"],
            }
        )


class QualitativeAnalysis(PopulationAnalysis):
    CLASS_COLUMN = "feature"

    def plot_analysis_compare(self, attribute: str, modality: str):
        """
        Comparison plot between reference data and simulation

        :param attribute:
        :param modality:

        :return: Plotly figure
        """

        analysis_table = self.analysis_table
        analysis_table = analysis_table.copy()[
            (analysis_table["attribute"] == attribute) & (analysis_table["modality"] == modality)
        ]

        fig = px.histogram(
            analysis_table,
            x=self.CLASS_COLUMN,
            y=list(self.populations.keys()) + [self.distributions_name],
            barmode="group",
            histnorm="percent",
        )

        # configure plot
        fig.update_layout(
            title=self.get_plot_title(attribute, modality),
            xaxis_title="Values",
            yaxis_title="Distribution (%)",
        )

        return fig

    def _format_distributions_for_analysis(self):
        feature_values = get_feature_from_qualitative_distribution(self.distributions)
        distributions = self.distributions[["attribute", "modality"] + feature_values]

        distributions_formated = distributions.melt(
            id_vars=["attribute", "modality"],
            value_vars=feature_values,
            value_name=self.VALUE_COLUMN,
            var_name=self.CLASS_COLUMN,
        )

        return distributions_formated

    def _compute_distribution(self, population: pd.DataFrame) -> pd.DataFrame:
        population.loc[:, "key"] = 1

        res = population.groupby(self.feature_column)["key"].agg("count")

        df = pd.DataFrame({self.CLASS_COLUMN: res.index, self.VALUE_COLUMN: res.values})
        df[self.VALUE_COLUMN] = df[self.VALUE_COLUMN] / df[self.VALUE_COLUMN].sum()

        return df



import logging as lg
import random
import pandas as pd
from bhepop2 import functions
from bhepop2.quantitative_enrichment import QuantitativeEnrichment


class QualitativeEnrichment(QuantitativeEnrichment):
    """
    A class for enriching population using entropy maximisation.

    Notations used in this class documentation:

    - :math:`M_{k}` : crossed modality k (combination of attribute modalities)
    - :math:`F_{i}` : feature value i
    """

    mode = "qualitative"

    #: json schema of the enrichment parameters Modified by POV
    parameters_schema = {
        "title": "QualitativeEnrichment parameters",
        "description": "Parameters of a population enrichment run",
        "type": "object",
        "properties": {},
    }

    def optimise(self):
        # compute crossed modalities frequencies
        self.log("Computing frequencies of crossed modalities", lg.INFO)
        self.crossed_modalities_frequencies = functions.compute_crossed_modalities_frequencies(
            self.population, self.modalities
        )
        self.log(
            "Number of crossed modalities present in the population: {}".format(
                len(self.crossed_modalities_frequencies)
            )
        )

        # compute vector of feature values
        self.log("Computing vector of all feature values", lg.INFO)
        self.feature_values = functions.get_feature_from_qualitative_distribution(
            self.distributions
        )
        self.nb_features = len(self.feature_values)
        self.log("Number of feature values: {}".format(self.nb_features))

        # compute matrix of constraints
        self.log("Computing optimization constraints", lg.INFO)
        self._compute_constraints()

        # run resolution
        self.log("Starting optimization by entropy maximisation", lg.INFO)
        self.optim_result = self._run_optimization()

        return self.optim_result

    # Modification by POV

    def _compute_feature_prob(self, attribute, modality):
        """
        Compute the probability of being in each feature interval with the given modality.

        :param attribute: attribute name
        :param modality: attribute modality

        :return: DataFrame
        """

        prob_df = self.distributions[
            self.distributions["modality"].isin([modality])
            & self.distributions["attribute"].isin([attribute])
        ]

        res = pd.DataFrame({"feature": self.feature_values})
        res["prob"] = res["feature"].apply(lambda x: prob_df[x])

        return res

    def assign_feature_value_to_pop(self):
        """
        Assign feature values to the population individuals using the algorithm results.

        :return: enriched population DataFrame
        """

        self.log("Drawing feature values for the population", lg.INFO)

        # compute the probability of being in each feature interval, for each crossed modality
        res = self._get_feature_probs()

        # associate each individual to a crossed modality
        self.crossed_modalities_frequencies["index"] = self.crossed_modalities_frequencies.index
        merge = self.population.merge(
            self.crossed_modalities_frequencies,
            how="left",
            on=functions.get_attributes(self.modalities),
        )

        # associate a feature value to the population individuals
        merge["feature"] = merge["index"].apply(lambda x: self._draw_feature(res, x))

        # remove irrelevant columns
        merge.drop(["index", "probability"], axis=1, inplace=True)

        return merge

    def _get_feature_probs(self):
        """
        For each crossed modality, compute the probability of belonging to a feature interval.

        Invert the crossed modality probabilities using Bayes.

        Compute

        .. math::

            P(f < F_{i} \mid M_{k}) = P(M_{k} \mid f < F_{i}) \\cdot \\frac{P(f < F_{i})}{P(M_{k})}

        :return: DataFrame
        """

        feature_probs = self._compute_feature_probabilities_from_distributions()

        res = self.optim_result

        nb_columns = len(res.columns)

        for c in res.columns:
            res[c] = res[c].map(lambda x: x[0])  # POV sur conseils de Valentin

        for i in range(nb_columns):
            res[i] = res[i] * feature_probs["prob"][i]

        for i in range(len(res)):
            total = res.iloc[i, :].sum()
            res.iloc[i, :] = res.iloc[i, :] / total

        pd.set_option("display.max_rows", 500)
        pd.set_option("display.max_columns", 500)
        pd.set_option("display.width", 1000)

        return res

    def _draw_feature(self, res, index):
        """
        Draw a feature value using the given distribution.

        :param res: feature distributions by crossed modality
        :param index: index of the crossed modality

        :return: Drawn feature value
        """

        # get probs
        probs = res.loc[index,].to_numpy()
        # interval_values = [self.parameters["abs_minimum"]] + self.feature_values POV

        # get the non-null probs and values
        probs2 = []
        values2 = []
        for i in range(len(probs)):
            if pd.isna(probs[i]):
                continue
            probs2.append(probs[i])
            values2.append(self.feature_values[i])
        #    values2.append(interval_values[i]) POV

        # TODO : probs don't sum to 1, this isn't reassuring

        # draw a feature interval using the probs
        # values = list(range(len(values2))) POV
        final = random.choices(values2, probs2)[0]
        # feature_interval = random.choices(values, probs2)[0]

        # draw a feature value using a uniform distribution in the interval POV
        # lower, upper = interval_values[feature_interval], interval_values[feature_interval + 1]
        # draw = random.random()
        # final = lower + round((upper - lower) * draw)

        return final

    def _compute_feature_probabilities_from_distributions(self):
        """
        Compute the probability of each feature interval.

        Use the global distribution of the features to interpolate the interval probabilities.

        The resulting DataFrame contains :math:`P(f < F_{i})` for i in N

        :return: DataFrame
        """
        distrib_all_df = self.distributions[self.distributions["attribute"] == "all"]

        assert len(distrib_all_df) == 1

        res = pd.DataFrame({"feature": self.feature_values})
        res["prob"] = res["feature"].apply(lambda x: distrib_all_df[x])

        return res



from bhepop2.quantitative_enrichment import QuantitativeEnrichment
from bhepop2.utils import *


def test_add_defaults_and_validate_against_schema(test_parameters):
    """
    Test the json schema validation.
    """

    add_defaults_and_validate_against_schema(
        test_parameters, QuantitativeEnrichment.parameters_schema
    )



from bhepop2.qualitative_enrichment import QualitativeEnrichment

import pytest
from bhepop2.functions import *


def build_cross_table(pop: pd.DataFrame, names_attribute: list):
    """


    Parameters
    ----------
    pop : DataFrame synthesis population
    names_attribute: list of two strings
           name of attribute1 and name of attribute 2

    Returns
    -------
    table_percentage : DataFrame
          proportion of modalities of attribute 2 given attribute 1


    """

    name_attribute1 = names_attribute[0]
    name_attribute2 = names_attribute[1]
    table_numbers = pd.crosstab(pop[name_attribute2], pop[name_attribute1])
    table_percentage_attribute2 = (
        table_numbers.transpose().sum() / table_numbers.transpose().sum().sum()
    )
    table_percentage = table_numbers / table_numbers.sum()
    table_percentage["all"] = table_percentage_attribute2
    table_percentage = table_percentage.transpose()
    table_percentage["modality"] = table_percentage.index
    table_percentage["attribute"] = name_attribute1

    return table_percentage


@pytest.fixture(scope="session")
def pop_synt_men_nantes():
    synt_pop = pd.read_csv("tests/data/pop_synth_men_nantes.csv")
    return synt_pop


@pytest.fixture(scope="session")
def synth_pop_defected(pop_synt_men_nantes):
    return pop_synt_men_nantes.drop(["Voit_rec"], axis=1)


@pytest.fixture(scope="session")
def distributions(pop_synt_men_nantes):
    attributes = list(pop_synt_men_nantes.columns[:-1])
    marginal_distribution = pd.concat(
        list(map(lambda a: build_cross_table(pop_synt_men_nantes, [a, "Voit_rec"]), attributes))
    )
    marginal_distribution = marginal_distribution.loc[
        ~marginal_distribution.index.duplicated(keep="first")
    ]
    marginal_distribution.loc["all", "attribute"] = "all"
    return marginal_distribution


def test_qualitative(synth_pop_defected, distributions, test_seed):
    enrich_class = QualitativeEnrichment(synth_pop_defected, distributions, seed=test_seed)
    # Run optimisation
    enrich_class.optimise()
    pop = enrich_class.assign_feature_value_to_pop()

    nb_cars_dict = pop.groupby("feature").agg("size").to_dict()
    assert nb_cars_dict["0voit"] == 88802
    assert nb_cars_dict["1voit"] == 299877
    assert nb_cars_dict["2voit"] == 206976
    assert nb_cars_dict["3voit"] == 33677
    assert sum(nb_cars_dict.values()) == len(pop)


# def test_qualitative_enrich(
#     pop_synt_men_nantes,
#     distributions,
#     test_modalities,
#     test_parameters,
#     test_seed,
#     expected_enriched_population_nantes,
# ):
#     modalities = infer_modalities_from_distributions(distributions)
#     modalities["Voit_rec"] = ["0voit", "1voit", "2voit", "3voit"]
#     test = compute_crossed_modalities_frequencies(pop_synt_men_nantes, modalities)
#
#
#     def calc(x):
#         return x["probability"] / distributions.loc["all", x["Voit_rec"]]
#
#     test["prob_cond"] = test.apply(calc, axis=1)
#
#     test.apply(
#     calc,
#     axis=1,
#     )
#
#     # le test commence ici
#     synt_pop_defected = pop_synt_men_nantes.drop(['Voit_rec'], axis=1)
#
#     enrich_class = QualitativeEnrichment(
#     synt_pop_defected, distributions, seed=test_seed
#     )
#
#     # Run optimisation
#     enrich_class.optimise()
#
#
#     # Assign data to synthetic population
#     pop = enrich_class.assign_feature_value_to_pop()
#     print(pop.head())
#     pop.rename(columns={"feature":"Voit_rec"},inplace=True) # ename feaure in Voit_rec in order to compare with the true population
#
#     # this solution will be compare with a naive resolution
#     # by samping according the general frequeny in the _init_population
#     # Récupérez les valeurs de 0voit ... 3voit pour la ligne où Voit_rec vaut 'all'
#
#     ligne_all_voit = distributions.loc[distributions['attribute'] == 'all', '0voit':'3voit'].values
#     ligne_all_voit = ligne_all_voit[0] # the output of the previous line is 2D, it is downsized to 1D
#     common_pop=synt_pop_defected
#     common_pop['Voit_rec']=np.random.choice(['0voit', '1voit', '2voit', '3voit'], size=len(common_pop), p=ligne_all_voit)
#     print(common_pop.head())
#
#     prediction = compute_crossed_modalities_frequencies(pop, modalities)
#     common_prediction = compute_crossed_modalities_frequencies(common_pop, modalities)
#
#     def build_probability_vector(probability_df, modalities, feature_column):
#         """
#         Inspecte les modalités croisées entre les colonnes du DataFrame probability_df et les modalités fournies
#         dans le dictionnaire modalities. Ajoute la valeur feature si la combinaison de modalités existe dans probability_df,
#         sinon ajoute zéro.
#
#         Args:
#         probability_df (pd.DataFrame): DataFrame à inspecter.
#         modalities (dict): Dictionnaire des modalités pour chaque colonne.
#         feature_column (str): Nom de la colonne contenant les valeurs de feature.
#
#         Returns:
#          np.array : vecteur des valeurs de feature ou zéro pour chaque combinaison de modalités.
#
#         Todo:
#           si on trouve plusieurs valeurs de probabilités pour la même modalité croisé, c'est une erreur
#         """
#         result = []
#         # Génère toutes les combinaisons possibles de modalités
#         combinations = list(itertools.product(*modalities.values()))
#
#         for combination in combinations:
#             # Vérifie si la combinaison de modalités existe dans pop
#             mask = probability_df.apply(lambda row: all(row[colname] in [modality] for colname, modality in zip(modalities.keys(), combination)), axis=1)
#             if mask.any():
#                 # Ajoute la valeur de feature si la combinaison de modalités existe
#                 result.append(probability_df.loc[mask, feature_column].values[0])
#             else:
#                 # Ajoute zéro si la combinaison de modalités n'existe pas
#                 result.append(0)
#         return np.array(result)
#
#     # Exemple d'utilisation
#     frequencies_vector = build_probability_vector(test, modalities, 'probability')
#     probabilities_vector = build_probability_vector(prediction, modalities, 'probability')
#     common_probabilities_vector = build_probability_vector(common_prediction, modalities, 'probability')
#
#     def compute_ratio(proba, freq, nagents, nmodalities, eps_proba, proba_min):
#         """
#         calcule les métriques entre les proba estimées et les fréquences réelles
#            RE : relative Error
#            R2 : coefficient of determination
#            RMSD : root Mean Square Deviation
#            MAPE : Mean Absolute Percentage Error
#            divergence : kullback leibler divergence
#            Goodness of fit :
#
#
#         Args:
#         proba : vecteur des probabilités estimés
#         freq : vecteur des fréquences réelles
#         nmodalities : nombre de modalités croisées
#         eps_proba : niveau pour détecter quand une probabilité est nulle
#         proba_min : une proba détectée comme nulle est mise à la valeur proba_min
#         eps : si v_input(k) < eps, il est considéré comme nul
#         proba_min : les probabilités nulles sont remplacés par la valeur proba_min
#
#         Returns:
#         rmsd, mape, divergence, goodness_of_fit
#
#         """
#
#         def support_vector_not_null(v_input, eps, proba_min):
#             """
#             construit un vecteur de probabilité qui ne contient pas probabilité nulle, ceci permet de comparer des vecteurs de probabilités qui n'ont pas le même support
#
#             Args:
#             v_input : vecteur de probabilité avec des composants nul
#             eps : si v_input(k) < eps, il est considéré comme nul
#             proba_min : les probabilités nulles sont remplacés par la valeur proba_min
#
#             Returns:
#             v_ouput: vecteur de probabilité dont aucune composante n'est nulles
#
#             TODO : la somme de v_output ne vaut plus 1... To discuss
#                     pas de test en entrée
#             """
#             v_output=v_input.copy()
#             v_output[v_input<eps]=proba_min
#             return v_output/np.sum(v_output)
#
#
#
#         freq_not_null = support_vector_not_null(freq,eps_proba,proba_min)
#         proba_not_null = support_vector_not_null(proba,eps_proba,proba_min)
#
#         diff_vector = proba - freq
#
#         re = np.linalg.norm(diff_vector)/np.linalg.norm(freq)
#         rmsd = np.linalg.norm(diff_vector)/np.sqrt(nmodalities)
#         r2 = 1-np.sum(diff_vector**2)/np.sum(freq**2)
#
#         diff_vector = proba - freq_not_null
#         mape = (np.sum(np.abs(diff_vector) / freq_not_null))/nmodalities
#
#         divergence = np.sum( freq_not_null*np.log(freq_not_null/proba_not_null))
#
#         diff_vector = proba_not_null - freq
#         goodness_of_fit = np.sum(diff_vector**2 / proba_not_null)
#
#         return re, r2, rmsd, mape, divergence, goodness_of_fit
#
#     nagents = len(pop)
#     nmodalities = len(frequencies_vector)
#     eps_proba = 1e-6/nagents
#     proba_min = 0.1/nagents # tobe discussed
#
#     re, r2, rmsd, map, divergence, goodness_of_fit = compute_ratio(probabilities_vector, frequencies_vector, nagents, nmodalities, eps_proba, proba_min)
#     c_re, c_r2, c_rmsd, c_map, c_divergence, c_goodness_of_fit = compute_ratio(common_probabilities_vector, frequencies_vector, nagents, nmodalities, eps_proba, proba_min)
#
#
#     # Display tests results
#     metric_bhepop2 = [re,r2,rmsd,map,divergence,goodness_of_fit]
#     metric_common = [c_re,c_r2,c_rmsd,c_map,c_divergence,c_goodness_of_fit]
#     criteria = ['Relative Error', 'Coef of determination', 'Root Mean Square Deviration', 'Mean Absolute Percentage Error', 'Divergence', 'Goodness of fit' ]
#     results_test = pd.DataFrame({'Common': metric_common, 'Bhepop2': metric_bhepop2 },index=criteria)
#     print(results_test)
#     assert 0 #POV pour rester dans l'environnement lorsque nous lançons pytest --pdb tests/test_qualitative.py -s



from bhepop2.uniform_enrich import uniform_enrich


def test_uniform_enrich(synthetic_population_nantes, filosofi_distributions_nantes, test_seed):
    filosofi_distributions_nantes = filosofi_distributions_nantes[
        filosofi_distributions_nantes["modality"] == "all"
    ]

    pop = uniform_enrich(
        synthetic_population_nantes, filosofi_distributions_nantes, 0, 1.2, test_seed
    )



import numpy as np
from numpy.linalg import norm
from .utils import log
import logging as lg


def minxent_gradient(
    q: np.ndarray, matrix: np.ndarray, eta: np.ndarray, lambda_: np.ndarray, maxiter: int
):
    # remove first constraint
    # TODO : check that first constraint is density constraint instead of discarding it
    matrix = matrix.copy()[1:, :]
    eta = eta.copy()[:, 1:]

    # get array size
    ncons, nproba = matrix.shape
    q = q.reshape(nproba, 1)
    lambda_ = lambda_.reshape(ncons, 1)
    eta = eta.reshape(ncons, 1)

    # remove ?
    Lagrangians = []

    # move to const ?
    common_ratio_descending = 0.5
    common_ratio_ascending = 2.0

    # loop until maximum iteration number or convergence
    iter_general = 0
    dist = 1
    while iter_general <= maxiter and dist >= 1e-08:
        iter_general += 1

        # lambda0 = np.log(np.sum(q * np.exp(-lambda_old.dot(G[1:, :])))) Exp_neg_Gt_lambdaPOV
        Exp_neg_Gt_lambda = np.exp(-matrix.T.dot(lambda_))  # POV exp(-Gt.lambda)
        lambda0 = np.log(q.T.dot(Exp_neg_Gt_lambda))
        pk = (q * Exp_neg_Gt_lambda) / (q.T.dot(Exp_neg_Gt_lambda))
        level_objective = lambda0 + np.sum(lambda_ * eta)
        f_old = eta - matrix.dot(pk)

        alpha_ascent = 1.0
        alpha_descent = 1.0
        alpha = 1.0
        lambda_old = lambda_.copy()
        lambda_ = lambda_.copy()
        did_ascent = False
        did_descent = False

        while not (did_ascent and did_descent):
            lambda_new = lambda_old - alpha * f_old
            lambda0 = np.log(q.T.dot(np.exp(-matrix.T.dot(lambda_new))))
            level_objective_new = lambda0 + np.sum(lambda_new * eta)

            # level_objective_new = objective(q=q, G=G, eta=eta, lambda_=lambda_new)
            if level_objective_new > level_objective:
                alpha_descent *= common_ratio_descending
                alpha = alpha_descent
                did_descent = True

            else:
                level_objective = level_objective_new
                lambda_ = lambda_new.copy()
                alpha_ascent *= common_ratio_ascending
                alpha = alpha_ascent
                did_ascent = True

            # fail to converge
            if alpha < 1e-06 or alpha > 1e6:
                log(
                    "Leaving gradient descent due to extreme alpha value : {}".format(alpha),
                    lg.DEBUG,
                )
                break

        dist = norm(lambda_ - lambda_old)

    # log("Leaving mixent after {} iterations".format(str(iter_general)), lg.DEBUG)
    Exp_neg_Gt_lambda = np.exp(-matrix.T.dot(lambda_))  # POV exp(-Gt.lambda)
    lambda0 = np.log(q.T.dot(Exp_neg_Gt_lambda))
    pk = (q * Exp_neg_Gt_lambda) / (q.T.dot(Exp_neg_Gt_lambda))

    # remove ?
    # pi_solve = q * np.exp(-lambda0) * np.exp(-lambda_.dot(G[1:, :]))
    Lagrangians.append([lambda0, lambda_.tolist()])
    test_pierreolivier = matrix.dot(pk) - eta
    # log("test PO : " + str(test_pierreolivier/eta), 10)
    return pk.tolist(), lambda_



import numpy as np


# TODO : refactor this function in a class similar to QuantitativeEnrichment (started below)
def uniform_enrich(population, distributions, abs_minimum, relative_maximum, seed):
    """

    :param population: synthetic population
    :param distributions: decile distribution (modality 'all')
    :param abs_minimum: see QuantitativeEnrichment parameters
    :param relative_maximum: see QuantitativeEnrichment parameters
    :param seed: random seed

    :return: enriched population
    """

    rng = np.random.RandomState(seed)

    centiles = list(
        distributions[["D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8", "D9"]].iloc[0].values
    )

    centiles = np.array([abs_minimum] + centiles + [np.max(centiles) * relative_maximum])

    pop_size = len(population)

    indices = rng.randint(10, size=pop_size)
    lower_bounds, upper_bounds = centiles[indices], centiles[indices + 1]

    incomes = lower_bounds + rng.random_sample(size=pop_size) * (upper_bounds - lower_bounds)

    population["feature"] = incomes
    population["feature"] = round(population["feature"])

    return population


# class UniformEnrich:
#
#     def __init__(self, population, distributions, abs_minimum, relative_maximum, seed):
#
#         # random generator
#         self.rng = np.random.RandomState(seed)
#
#         # original population to be enriched
#         self.population = None
#
#         # distributions of the feature values by modality
#         self.distributions = None
#
#         self.abs_minimum = abs_minimum
#         self.relative_maximum = relative_maximum
#
#         # algorithm data
#
#         # vector of feature values defining the intervals
#         self.feature_values = None
#
#         # total number of features
#         self.nb_features = None
#
#         self.result_probs = None
#
#         self.log("Initialisation of enrichment algorithm data", lg.INFO)
#
#         self.log("Setup distributions data")
#         self._init_distributions(distributions)
#
#         self.log("Setup population data")
#         self._init_population(population)
#
#     def _init_distributions(self, distributions):
#         self.distributions = distributions.copy()
#
#         # TODO : assert only one line with attribute all
#
#     def _init_population(self, population):
#         self.population = population.copy()
#
#     def evaluate_probs(self):
#
#         centiles = list(self.distributions[
#                             ["D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8", "D9"]].iloc[0].values)
#
#         centiles = np.array([self.abs_minimum] + centiles + [np.max(centiles) * self.relative_maximum])
#
#         self.feature_values = centiles
#
#     def draw_feature_values(self):
#
#         pop_size = len(self.population)
#
#         indices = self.rng.randint(10, size=pop_size)
#         lower_bounds, upper_bounds = self.feature_values[indices], self.feature_values[indices + 1]
#
#         incomes = lower_bounds + self.rng.random_sample(size=pop_size) * (upper_bounds - lower_bounds)
#
#         self.population["feature"] = incomes
#         self.population["feature"] = round(self.population["feature"])
#
#         return self.population
#
#     def log(message: str, level: int = lg.DEBUG):
#         """
#         Log a message using the package logger.
#
#         See logging library.
#
#         :param message: message to be logged
#         :param level: logging level
#         """
#
#         utils.log(message, level)
#
#     log = staticmethod(log)



"""
Utility functions
"""
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


## functions for processing synthetic populations


def add_household_size_attribute(
    population: pd.DataFrame,
    values_map: callable = lambda x: str(x) + "_pers" if x < 5 else "5_pers_or_more",
    person_id: str = "person_id",
    household_id: str = "household_id",
    column_name: str = "size",
):
    """
    Add a size attribute to the given synthetic population.

    Even though we add the attribute at a person's level,
    its value is related to the household (we add household size).

    :param population: synthetic population df
    :param values_map: mapping function applied to household count
    :param person_id: name of the column containing persons' ids
    :param household_id: name of the column containing households' ids
    :param column_name: name of the added column

    :return: population with new 'column_name' column containing household size
    """

    # make copy of original population
    population = population.copy()

    # count number of persons by household
    df_households_size = population.groupby([household_id], as_index=False)[person_id].count()

    # rename size column
    df_households_size.rename(columns={person_id: column_name}, inplace=True)

    # map size values if a mapper is provided
    if values_map is not None:
        df_households_size[column_name] = df_households_size[column_name].apply(values_map)

    # add size column to population table
    population = pd.merge(population, df_households_size, on="household_id")

    return population


def add_household_type_attribute(
    population,
    person_id: str = "person_id",
    household_id: str = "household_id",
    column_name="family_comp",
):
    """
    Add a type attribute to the given synthetic population.

    The following attributes are needed on the population:
        - household_size: number of person's in the household
        - sex: person's sex
        - age: person's age
        - couple: boolean indicating if the person is in couple in the household

    The household type can take the following values:
        - Single_man
        - Single_wom
        - Couple_without_child
        - Couple_with_child
        - Single_parent
        - complex_hh

    Even though we add the attribute at a person's level,
    its value is related to the household (we add household type).

    :param population: synthetic population df
    :param person_id: name of the column containing persons' ids
    :param household_id: name of the column containing households' ids
    :param column_name: name of the added column

    :return: population with new 'column_name' column containing household size
    """

    # make copy of original population
    population = population.copy()

    # get single men and women
    df_households_info = population.groupby([household_id], as_index=False).agg(
        {"household_size": "first", "sex": "first"}
    )
    df_households_single_man = df_households_info.query(
        "household_size==1 and sex=='male'"
    ).reset_index()
    df_households_single_man[column_name] = "Single_man"
    df_households_single_woman = df_households_info.query(
        "household_size==1 and sex=='female'"
    ).reset_index()
    df_households_single_woman[column_name] = "Single_wom"

    # get couple without child
    # 2 persons and with couple=True
    df_households_couple = (
        population.query("couple==True")
        .groupby([household_id], as_index=False)
        .agg({"couple": "count", "household_size": "first"})
        .query("couple==2")
    )
    df_households_couple_without_child = df_households_couple.query(
        "couple==2 and household_size==2"
    ).reset_index()
    df_households_couple_without_child[column_name] = "Couple_without_child"

    # get couple with child
    # 3 persons or more, 2 persons in couple, and max age of person non couple < 25
    df_households_child = (
        population.query("age<25 and couple==False")
        .groupby([household_id], as_index=False)
        .agg({person_id: "count"})
        .rename(columns={person_id: "child_count"})
    )
    df_households_couple_with_child = (
        df_households_couple.merge(df_households_child, how="left", on=household_id)
        .query("household_size==child_count+2")
        .reset_index()
    )
    df_households_couple_with_child[column_name] = "Couple_with_child"

    # get single parent
    # 2 persons or more, no one in couple, oldest person = [25-60[, and others age < 25
    df_households_no_couple = population.query("couple==False and household_size>=2").sort_values(
        [household_id, "age"], ascending=[True, False]
    )

    df_households_no_couple_oldest = (
        df_households_no_couple.groupby([household_id], as_index=False)
        .agg({"age": "first", person_id: "first"})
        .rename(columns={"age": "oldest_age", person_id: "oldest_person_id"})
    )

    df_households_no_couple_second_oldest = (
        df_households_no_couple.merge(df_households_no_couple_oldest, how="left")
        .query(f"{person_id}!=oldest_person_id")
        .sort_values([household_id, "age"], ascending=[True, False])
        .groupby([household_id], as_index=False)
        .agg({"age": "first"})
        .rename(columns={"age": "second_oldest_age"})
    )

    df_households_single_parent = df_households_no_couple_oldest.merge(
        df_households_no_couple_second_oldest, how="left", on=household_id
    )
    df_households_single_parent = df_households_single_parent.query(
        "oldest_age>=25 and oldest_age<60 and second_oldest_age<25"
    ).reset_index()
    df_households_single_parent[column_name] = "Single_parent"

    # combine type
    combined = pd.concat(
        [
            df_households_single_man,
            df_households_single_woman,
            df_households_couple_without_child,
            df_households_couple_with_child,
            df_households_single_parent,
        ]
    )[[household_id, column_name]]

    assert combined[household_id].is_unique

    population = population.merge(combined, how="left", on=household_id)
    population[column_name].fillna("complex_hh", inplace=True)

    return population


## functions for reading Filosofi data

filosofi_attributes = [
    {
        "name": "all",
        "modalities": [
            {"name": "all", "sheet": "ENSEMBLE", "col_pattern": ""},
        ],
    },
    {
        "name": "size",
        "modalities": [
            {"name": "1_pers", "sheet": "TAILLEM_1", "col_pattern": "TME1"},
            {"name": "2_pers", "sheet": "TAILLEM_2", "col_pattern": "TME2"},
            {"name": "3_pers", "sheet": "TAILLEM_3", "col_pattern": "TME3"},
            {"name": "4_pers", "sheet": "TAILLEM_4", "col_pattern": "TME4"},
            {"name": "5_pers_or_more", "sheet": "TAILLEM_5", "col_pattern": "TME5"},
        ],
    },
    {
        "name": "family_comp",
        "modalities": [
            {"name": "Single_man", "sheet": "TYPMENR_1", "col_pattern": "TYM1"},
            {"name": "Single_wom", "sheet": "TYPMENR_2", "col_pattern": "TYM2"},
            {"name": "Couple_without_child", "sheet": "TYPMENR_3", "col_pattern": "TYM3"},
            {"name": "Couple_with_child", "sheet": "TYPMENR_4", "col_pattern": "TYM4"},
            {"name": "Single_parent", "sheet": "TYPMENR_5", "col_pattern": "TYM5"},
            {"name": "complex_hh", "sheet": "TYPMENR_6", "col_pattern": "TYM6"},
        ],
    },
    {
        "name": "age",
        "modalities": [
            {"name": "0_29", "sheet": "TRAGERF_1", "col_pattern": "AGE1"},
            {"name": "30_39", "sheet": "TRAGERF_2", "col_pattern": "AGE2"},
            {"name": "40_49", "sheet": "TRAGERF_3", "col_pattern": "AGE3"},
            {"name": "50_59", "sheet": "TRAGERF_4", "col_pattern": "AGE4"},
            {"name": "60_74", "sheet": "TRAGERF_5", "col_pattern": "AGE5"},
            {"name": "75_or_more", "sheet": "TRAGERF_6", "col_pattern": "AGE6"},
        ],
    },
    {
        "name": "ownership",
        "modalities": [
            {"name": "Owner", "sheet": "OCCTYPR_1", "col_pattern": "TOL1"},
            {"name": "Tenant", "sheet": "OCCTYPR_2", "col_pattern": "TOL2"},
        ],
    },
    {
        "name": "income_source",
        "modalities": [
            {"name": "Salary", "sheet": "OPRDEC_1", "col_pattern": "OPR1"},
            {"name": "Unemployment", "sheet": "OPRDEC_2", "col_pattern": "OPR2"},
            {"name": "Independent", "sheet": "OPRDEC_3", "col_pattern": "OPR3"},
            {"name": "Pension", "sheet": "OPRDEC_4", "col_pattern": "OPR4"},
            {"name": "Property", "sheet": "OPRDEC_5", "col_pattern": "OPR5"},
            {"name": "None", "sheet": "OPRDEC_6", "col_pattern": "OPR6"},
        ],
    },
]


def read_filosofi(filepath: str, year: str, attributes: list, communes=None):
    """
    Fetch income distributions by attribute (age, sex, ...) and commune from Filosofi file

    :param filepath: path to Filosofi excel file (DISP_COM)
    :param year: Filosofi data year
    :param attributes: list of attributes and their modalities, with data required for extraction
    :param communes: optional list of communes to filter

    :return: distributions DataFrame
    """
    # build full list of sheets
    sheet_list = []
    for attribute in attributes:
        sheet_list = sheet_list + [x["sheet"] for x in attribute["modalities"]]

    # read Filosofi excel file
    filosofi_sheets = read_filosofi_excel(filepath, sheet_list)

    # fetch distributions for the given attributes
    distributions = read_filosofi_attributes(filosofi_sheets, year, attributes, communes)

    return distributions


def read_filosofi_excel(filepath: str, sheet_list: list):
    """
    Read list of sheets from Filosofi excel file.

    :param filepath: path to Filosofi excel file (DISP_COM)
    :param sheet_list: list of sheets to be read

    :return: DataFrame indexed by sheet
    """
    excel_df = pd.read_excel(filepath, sheet_name=sheet_list, skiprows=5)
    return excel_df


def read_filosofi_attributes(filosofi_sheets, year, attributes: list, communes=None):
    """
    Read distributions from list of attributes and their modalities in filosofi sheets.

    :param filosofi_sheets: Filosofi excel as DataFrame indexed by sheet
    :param year: Filosofi data year
    :param attributes: list of attributes and their modalities, with data required for extraction
    :param communes: optional list of communes to filter

    :return: distributions DataFrame
    """

    concat_list = []

    # browse modalities for each attribute
    for attribute in attributes:
        for modality in attribute["modalities"]:
            # read distributions from filosofi data
            data = read_distributions_from_filosofi(
                filosofi_sheets,
                year,
                modality["sheet"],
                modality["col_pattern"],
                attribute["name"],
                modality["name"],
                communes,
            )

            concat_list.append(data)

    df = pd.concat(concat_list)

    # Validation
    assert len(attributes) == len(df["attribute"].unique())

    return df[
        [
            "commune_id",
            "D1",
            "D2",
            "D3",
            "D4",
            "D5",
            "D6",
            "D7",
            "D8",
            "D9",
            "reference_median",
            "attribute",
            "modality",
        ]
    ]


def read_distributions_from_filosofi(
    filosofi_sheets,
    year: str,
    sheet: str,
    col_pattern: str,
    attribute: str,
    modality: str,
    communes=None,
):
    """


    :param filosofi_sheets: Filosofi excel as DataFrame indexed by sheet
    :param year: Filosofi data year
    :param sheet: sheet name
    :param col_pattern: column pattern
    :param attribute: attribute name
    :param modality: modality name
    :param communes: optional list of communes to filter

    :return: distributions DataFrame
    """

    # Load income distribution
    data = filosofi_sheets[sheet][
        ["CODGEO"]
        + [
            "%sD%d" % (col_pattern, q) + year if q != 5 else col_pattern + "Q2" + year
            for q in range(1, 10)
        ]
    ].copy()
    data.columns = ["commune_id", "D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8", "D9"]
    data.loc[:, "reference_median"] = data["D5"]
    data.loc[:, "modality"] = modality
    data.loc[:, "attribute"] = attribute

    if communes is not None:
        data = data[data["commune_id"].isin(communes)]

    return data



"""
Utility functions
"""

import pandas as pd
import logging as lg
from jsonschema import validate, ValidationError

# log utils (see logging library)

#: logging level
logger_level = lg.WARNING

#: logger name
logger_name = "bhepop2_logger"


def log(message, level):
    """
    Log a message using the logging library.

    :param message: message to log
    :param level: logging level
    """

    # get logger object
    logger = _get_logger()

    # log message
    logger.log(level, message)


def _get_logger():
    """
    Create a logger or return the current one if already instantiated.

    :return: logging.logger
    """

    logger = lg.getLogger(logger_name)

    # if a logger with this name is not already set up
    if not getattr(logger, "handler_set", None):
        logger.propagate = False
        formatter = lg.Formatter("%(message)s")
        stream_handler = lg.StreamHandler()
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
        logger.setLevel(logger_level)
        logger.handler_set = True

    return logger


# json schema utils


def add_defaults_and_validate_against_schema(instance, schema):
    """
    Add default values then validate instance against the schema.

    :param instance: data instance
    :param schema: json schema

    :return: result data (copy of instance)
    """

    result = instance.copy()

    # superficial default values set
    for key in schema["properties"]:
        if "default" in schema["properties"][key] and not key in result:
            result[key] = schema["properties"][key]["default"]

    # validate instance against schema
    try:
        validate(result, schema)
    except ValidationError as e:
        msg = e.message
        raise ValueError(msg) from None

    return result



import pandas as pd
import numpy as np
from scipy.optimize import linprog


# generic functions


def get_attributes(modalities: dict) -> list:
    """
    Get attributes list from dictionary of modalities

    :param modalities:

    :return: attributes
    """
    return list(modalities.keys())


def modality_feature(attribute, modality) -> callable:
    """
    Create a function that checks if a sample belongs to the given attribute and modality.

    :param attribute: attribute value
    :param modality: modality value

    :return: feature checking function
    """

    def feature(x):
        return x[attribute] == modality

    return feature


# distribution functions


def validate_distributions(distributions: pd.DataFrame, attribute_selection, mode):
    """
    Validate the format and contents of the given distribution.

    :param distributions: distribution DataFrame
    :param attribute_selection: list of attributes to keep in the distribution, or None
    :param mode: "qualitative" or "quantitative"
    :raises: AssertionError
    """

    assert not distributions.empty, "Empty distributions table provided"

    if mode == "quantitative":
        # we could validate the distributions columns (positive, monotony ?)
        assert {*["D{}".format(i) for i in range(1, 10)], "attribute", "modality"} <= set(
            distributions.columns
        ), "Distributions table lacks the required columns"
    elif mode == "qualitative":
        assert "attribute" in distributions.columns and "modality" in distributions.columns
    else:
        raise ValueError(f"Unknown mode '{mode}'")

    if attribute_selection is not None:
        # check that the distributions contain the selected attributes
        assert set(attribute_selection + ["all"]) <= set(
            distributions["attribute"]
        ), "Distributions table does not include selected attributes"


def filter_distributions_and_infer_modalities(distributions: pd.DataFrame, attribute_selection):
    """
    Filter distributions table with attribute selection and infer modalities.

    :param distributions: distribution DataFrame
    :param attribute_selection: list of attributes to keep in the distribution, or None

    :return: filtered distribution Dataframe, { attribute: [modalities] } dict
    """
    # make a copy of the distributions
    distributions = distributions.copy()

    if attribute_selection is not None:
        # filter distributions attributes
        distributions = distributions[
            distributions["attribute"].isin(attribute_selection + ["all"])
        ]

    # infer attributes and their modalities from the filtered distribution
    modalities = infer_modalities_from_distributions(distributions)

    return distributions, modalities


def infer_modalities_from_distributions(distributions: pd.DataFrame):
    """
    Infer attributes and their modalities from the given distributions.

    :param distributions: distributions DataFrame

    :return: dict of attributes and their modalities, { attribute: [modalities] }
    """

    # group by attribute and get all modality values
    modalities = distributions.groupby("attribute")["modality"].apply(list).to_dict()

    # remove global distribution
    if "all" in modalities:
        del modalities["all"]

    return modalities


# TODO : rename with reference to quantitative nature of distribution. Or move to quantitative class
def compute_feature_values(
    distribution: pd.DataFrame, relative_maximum: float, delta_min=None
) -> list:
    """
    Compute the list of feature values that will define the assignment intervals.

    The distributions do not give the knowledge of the minimum and maximum
    feature values, so we have to choose them.
    The minimum is the same for all distributions, it is directly equal to the abs_first_value parameter.
    The maximum is computed by multiplying the relative_maximum parameter to the last value of each distribution.

    :param distribution: dataframe of distribution
    :param relative_maximum: multiplicand applied to compute the last feature value of each distribution
    :param delta_min: minimum delta between two feature values. None to keep all values.

    :return: list of feature values
    """
    deciles = distribution.copy()

    # set maximum values
    deciles["D10"] = deciles["D9"] * relative_maximum

    # restraint columns to distribution values and get vector
    deciles = deciles[list(map(lambda a: "D" + str(a), list(range(1, 11))))]
    vec_all = list(deciles.to_numpy().flatten())

    # sort values vector
    vec_all.sort()

    # remove close values using delta min
    if delta_min is not None:
        last_value = vec_all[0]
        filtered_vec = [last_value]
        for val in vec_all[1:]:
            if val - last_value >= delta_min:
                filtered_vec.append(val)
                last_value = val
        vec_all = filtered_vec

    return vec_all


def get_feature_from_qualitative_distribution(distribution: pd.DataFrame):
    """
    Get feature values from the given distributions.

    :param distribution: distribution DataFrame

    :return: list of possible values for the qualitative feature
    """

    features = list(distribution.columns)
    features.remove("attribute")
    features.remove("modality")

    assert (distribution[features].apply(lambda row: np.isclose(row.sum(), 1), axis=1)).all()

    return features


def compute_features_prob_qualitative(feature_values: list, distribution: list):
    """
    Create a DataFrame containing probabilities for the given feature values.

    :param feature_values: list of feature values
    :param distribution: list of distribution values

    :return: DataFrame of feature probabilities
    """
    # set features column
    probs_df = pd.DataFrame({"feature": feature_values})

    # compute prob of being in each feature interval
    probs_df["prob"] = probs_df.apply(
        lambda x: interpolate_feature_prob(x["feature"], distribution),
        axis=1,
    )


def compute_features_prob(feature_values: list, distribution: list):
    """
    Create a DataFrame containing probabilities for the given feature values.

    :param feature_values: list of feature values
    :param distribution: list of distribution values

    :return: DataFrame of feature probabilities
    """
    # set features column
    probs_df = pd.DataFrame({"feature": feature_values})

    # compute prob of being in each feature interval
    probs_df["cumulative"] = probs_df.apply(
        lambda x: interpolate_feature_prob(x["feature"], distribution),
        axis=1,
    )

    probs_df["tmp"] = probs_df["cumulative"].shift(1)
    probs_df = probs_df.fillna(0)
    probs_df["prob"] = probs_df["cumulative"] - probs_df["tmp"]
    probs_df.drop(columns=["cumulative", "tmp"], inplace=True)

    return probs_df


def interpolate_feature_prob(feature_value: float, distribution: list):
    """
    Linear interpolation of a feature value probability.

    :param feature_value: value of feature to interpolate
    :param distribution: feature values for each decile from 1 to 10 (without value for 0)

    :return: probability of being lower than the input feature value
    """
    distribution = [0] + distribution
    if feature_value > distribution[10]:
        return 1
    if feature_value < distribution[0]:
        return 0
    decile_top = 0
    while feature_value > distribution[decile_top]:
        decile_top += 1

    interpolation = (feature_value - distribution[decile_top - 1]) * (
        decile_top * 0.1 - (decile_top - 1) * 0.1
    ) / (distribution[decile_top] - distribution[decile_top - 1]) + (decile_top - 1) * 0.1

    return interpolation


# population functions


def validate_population(population: pd.DataFrame, modalities: dict):
    """
    Validate the format and contents of the given population.

    Check that the population is compatible with the chosen modalities.

    :param population: distribution DataFrame
    :param modalities:
    :raises: AssertionError
    """

    attributes = get_attributes(modalities)

    assert {*attributes} <= set(population.columns)

    for attribute in attributes:
        assert population[attribute].isin(modalities[attribute]).all(), (
            f"Population validation: one of the modality values was not "
            f"found in distributions for the attribute '{attribute}'"
        )


def compute_crossed_modalities_frequencies(
    population: pd.DataFrame, modalities: dict
) -> pd.DataFrame:
    """
    Compute the frequency of each crossed modality present in the population.

    Columns other than attributes are removed from the result DataFrame, and a
    'probability' column is added.

    :param population: population DataFrame
    :param modalities: modalities dict

    :return: DataFrame of crossed modalities frequencies
    """

    attributes = get_attributes(modalities)

    # group by attributes and count the number of individuals, then divide by total
    population["count"] = 1
    freq_df = population.groupby(attributes, as_index=False)["count"].sum()
    freq_df["probability"] = freq_df["count"] / sum(freq_df["count"])
    freq_df = freq_df[attributes + ["probability"]]

    # remove count column
    population.drop("count", axis=1, inplace=True)

    return freq_df


# check functions


def compute_rq(model, nb_modalities, K):
    I = np.identity(nb_modalities)
    A = model.F.A
    A_eq = np.concatenate([A, I], axis=1)

    c = np.concatenate([np.zeros(np.shape(A)[1]), np.ones(nb_modalities)], axis=None)

    res = linprog(c, A_eq=A_eq, b_eq=K, method="simplex")

    return res



from bhepop2.functions import infer_modalities_from_distributions
from bhepop2.analysis import QualitativeAnalysis, QuantitativeAnalysis

import pandas as pd


def build_cross_table(pop: pd.DataFrame, names_attribute: list):
    """


    Parameters
    ----------
    pop : DataFrame synthesis population
    names_attribute: list of two strings
           name of attribute1 and name of attribute 2

    Returns
    -------
    table_percentage : DataFrame
          proportion of modalities of attribute 2 given attribute 1


    """

    name_attribute1 = names_attribute[0]
    name_attribute2 = names_attribute[1]
    table_numbers = pd.crosstab(pop[name_attribute2], pop[name_attribute1])
    table_percentage_attribute2 = (
        table_numbers.transpose().sum() / table_numbers.transpose().sum().sum()
    )
    table_percentage = table_numbers / table_numbers.sum()
    table_percentage["all"] = table_percentage_attribute2
    table_percentage = table_percentage.transpose()
    table_percentage["modality"] = table_percentage.index
    table_percentage["attribute"] = name_attribute1

    return table_percentage


def test_quantitative_analysis(
    filosofi_distributions_nantes, expected_enriched_population_nantes, test_modalities, tmp_dir
):
    analysis = QuantitativeAnalysis(
        {"qualitative": expected_enriched_population_nantes},
        test_modalities,
        "feature",
        filosofi_distributions_nantes,
        output_folder=tmp_dir,
    )

    analysis.generate_analysis_plots()
    analysis.generate_analysis_error_table()


def test_qualitative_analysis(pop_synt_men_nantes, tmp_dir):
    # synth pop
    pop = pop_synt_men_nantes.copy()

    # distributions
    attributes = list(pop.columns[:-1])
    marginal_distribution = pd.concat(
        list(map(lambda a: build_cross_table(pop, [a, "Voit_rec"]), attributes))
    )
    # removing multiple 'all'
    marginal_distribution = marginal_distribution.loc[
        ~marginal_distribution.index.duplicated(keep="first")
    ]
    marginal_distribution.loc["all", "attribute"] = "all"
    marginal_distribution = marginal_distribution.reset_index(drop=True)

    modalities = infer_modalities_from_distributions(marginal_distribution)

    analysis = QualitativeAnalysis(
        {"qualitative": pop}, modalities, "Voit_rec", marginal_distribution, output_folder=tmp_dir
    )

    analysis.generate_analysis_plots()



"""
Starling setup script.

See license in LICENSE.txt.
"""

import setuptools
import os
from bhepop2 import __version__

# short description of the project
DESC = "Synthetic population enrichment from aggregated data"

# get long description from README.md
LONG_DESC = """
A common problem in **generating a representative synthetic population** is that not all attributes of interest are present in the sample.  The purpose is to enrich the synthetic population with additional attributes, after the synthetic population is generated from the original sample.
In many cases, practitioners only have access to aggregated data for important socio-demographic attributes, such as income, education level. 
This package treats the problem to **enrich an initial synthetic population from an aggregated data** provided in the form of a distribution like deciles or quartiles.

Read the [docs](https://bhepop2.readthedocs.io/en/latest/) or see the code and examples on [GitHub](https://github.com/tellae/bhepop2).
"""

# list of classifiers from the PyPI classifiers trove
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "Topic :: Scientific/Engineering",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "License :: CeCILL-B Free Software License Agreement (CECILL-B)",
]

# only specify install_requires if not in RTD environment
if os.getenv("READTHEDOCS") == "True":
    INSTALL_REQUIRES = []
else:
    with open("requirements.txt") as f:
        INSTALL_REQUIRES = [line.strip() for line in f.readlines()]

# call setup
setuptools.setup(
    name="bhepop2",
    version=__version__,
    license="CECILL-B",
    author="UGE & Tellae",
    author_email="contact@tellae.fr",
    description=DESC,
    long_description_content_type="text/markdown",
    long_description=LONG_DESC,
    url="https://github.com/tellae/bhepop2",
    packages=setuptools.find_packages(),
    classifiers=CLASSIFIERS,
    python_requires=">=3.9",
    install_requires=INSTALL_REQUIRES,
)



import logging as lg
import numpy as np
import random
import pandas as pd
from bhepop2 import utils
from bhepop2 import functions
from .optim import minxent_gradient


class QuantitativeEnrichment:
    """
    A class for enriching population using entropy maximisation.

    Notations used in this class documentation:

    - :math:`M_{k}` : crossed modality k (combination of attribute modalities)

    - :math:`F_{i}` : feature class i

        - For quantitative features, corresponds to a numeric interval.
        - For qualitative features, corresponds to one of the feature values.
    """

    mode = "quantitative"

    #: json schema of the enrichment parameters
    parameters_schema = {
        "title": "Enrichment parameters",
        "description": "Parameters of a population enrichment run",
        "type": "object",
        "required": [
            "abs_minimum",
            "relative_maximum",
        ],
        "properties": {
            "abs_minimum": {
                "title": "Distributions absolute minimum",
                "description": "Minimum value of the feature distributions. This value is absolute, and thus equal for all distributions.",
                "type": "number",
                "default": 0,
            },
            "relative_maximum": {
                "title": "Distributions relative maximum",
                "description": "Maximum value of the feature distributions. This value is relative and will be multiplied to the last value of each distribution.",
                "type": "number",
                "default": 1.5,
                "minimum": 1,
            },
            "delta_min": {
                "title": "Minimum feature value delta",
                "description": "Minimum size of the feature intervals",
                "type": ["null", "number"],
                "default": None,
                "minimum": 0,
            },
        },
    }

    def __init__(
        self,
        population: pd.DataFrame,
        distributions: pd.DataFrame,
        attribute_selection: list = None,
        parameters=None,
        seed=None,
    ):
        """
        Synthetic population enrichment class.

        :param population: enriched population
        :param distributions: enriching data distributions
        :param attribute_selection: distribution attributes used. By default, use all attributes of the distribution
        :param parameters: enrichment parameters
        :param seed: random seed
        """

        # random seed (maybe use a random generator instead)
        self.seed = seed
        if seed is not None:
            random.seed(seed)

        # original population to be enriched
        self.population = None

        # distributions of the feature values by modality
        self.distributions = None

        # attributes considered for the assignment, with their modalities
        # { attribute: [modalities] }
        self.modalities = None

        # execution parameters
        if parameters is None:
            parameters = dict()
        self.parameters = utils.add_defaults_and_validate_against_schema(
            parameters, self.parameters_schema
        )

        # algorithm data

        # vector of values defining the feature classes
        self.feature_values = None

        # total number of features
        self.nb_features = None

        # frequency of each crossed modality present in the population
        self.crossed_modalities_frequencies = None

        # crossed modalities matrix
        self.crossed_modalities_matrix = None

        # optimization constraints
        self.constraints = None

        # optimization result
        self.optim_result = None

        self.log("Initialisation of enrichment algorithm data", lg.INFO)

        self._init_distributions(distributions, attribute_selection)
        self._init_population(population)

    def _init_distributions(self, distributions, attribute_selection):
        """
        Validate and filter the input distributions.

        When done, set the *distributions* field.

        :param distributions: input distributions DataFrame
        :param attribute_selection: distribution attributes selection
        """

        self.log("Setup distributions data")

        # validate distributions format and contents
        functions.validate_distributions(distributions, attribute_selection, self.mode)

        # filter distributions and infer modalities
        self.distributions, self.modalities = functions.filter_distributions_and_infer_modalities(
            distributions, attribute_selection
        )

        # check that there are modalities at the end
        assert (
            len(self.modalities.keys()) > 0
        ), "No attributes found in distributions for enriching population"

    def _init_population(self, population):
        """
        Validate and filter the input population.

        When done, set the *population* field.

        :param population: input population DataFrame
        """
        self.log("Setup population data")

        functions.validate_population(population, self.modalities)

        self.population = population.copy()

        # TODO ? remove distributions unused by population

    def optimise(self):
        """
        Run the optimisation algorithm to find the probability distributions that maximise entropy.

        When done, set the *optim_result* attribute.
        """
        # compute crossed modalities frequencies
        self.log("Computing frequencies of crossed modalities", lg.INFO)
        self.crossed_modalities_frequencies = functions.compute_crossed_modalities_frequencies(
            self.population, self.modalities
        )
        self.log(
            "Number of crossed modalities present in the population: {}".format(
                len(self.crossed_modalities_frequencies)
            )
        )

        # compute vector of feature values
        self.log("Computing vector of all feature values", lg.INFO)
        self.feature_values = functions.compute_feature_values(
            self.distributions, self.parameters["relative_maximum"], self.parameters["delta_min"]
        )
        self.nb_features = len(self.feature_values)
        self.log("Number of feature values: {}".format(self.nb_features))

        # compute matrix of constraints
        self.log("Computing optimization constraints", lg.INFO)
        self._compute_constraints()

        # run resolution
        self.log("Starting optimization by entropy maximisation", lg.INFO)
        self.optim_result = self._run_optimization()

        return self.optim_result

    def _run_optimization(self) -> pd.DataFrame:
        """
        Run optimization model on each feature value.

        The resulting probabilities are the :math:`P(M_{k} \\mid f \\in F_{i})`.

        :return: DataFrame containing the result probabilities
        """

        self.crossed_modalities_matrix = self._compute_crossed_modalities_matrix()
        nb_lines, nb_cols = self.crossed_modalities_matrix.shape

        res = pd.DataFrame()
        lambda_ = np.zeros(nb_lines - 1)

        # loop on features
        for i in range(self.nb_features):
            # run optimization model on the current feature
            self.log("Running optimization algorithm on feature " + str(i))

            q = self.crossed_modalities_frequencies["probability"].values.copy()
            ######### Change to start a prior uniform  tested but it does not change a lot the results
            # dimension = len(q)  # Dimension du vecteur
            # value = 1 / dimension  # Valeur pour toutes les composantes
            # q = np.full(dimension, value)  # Création du vecteur

            k = [1]
            for attribute in self.modalities:
                for modality in self.modalities[attribute][:-1]:
                    k.append(self.constraints[attribute][modality][i])
            k = np.array([k])

            res.loc[:, i], lambda_ = minxent_gradient(
                q=q, matrix=self.crossed_modalities_matrix, eta=k, lambda_=lambda_, maxiter=1000
            )

        return res

    def _compute_crossed_modalities_matrix(self):
        """
        Compute crossed modalities matrix for the present modalities.

        A reducted samplespace is evaluated from the crossed modalities present in the
        population. Functions describing each modality are then applied to elements of
        this samplespace.

        For each modality m and sample c, M(m, c) is 1 if c has modality m, 0 otherwise.

        :return: crossed_modalities_matrix describing crossed modalities
        """

        # samplespace is the set of all possible combinations
        attributes = functions.get_attributes(self.modalities)

        # get samplespace
        samplespace_reducted = self.crossed_modalities_frequencies[attributes].to_dict(
            orient="records"
        )

        features = []

        # base feature is x in samplespace
        def f0(x):
            return x in samplespace_reducted

        features.append(f0)

        # add a feature for all modalities except one for all variables
        for attribute in attributes:
            for modality in self.modalities[attribute][:-1]:
                features.append(functions.modality_feature(attribute, modality))

        nb_lines = len(features)
        nb_cols = len(samplespace_reducted)

        crossed_modalities_matrix = np.zeros((nb_lines, nb_cols))
        for i, f_i in enumerate(features):
            for j in range(nb_cols):
                f_i_x = f_i(samplespace_reducted[j])
                if f_i_x != 0:
                    crossed_modalities_matrix[i, j] = f_i_x

        return crossed_modalities_matrix

    def _compute_constraints(self):
        """
        For each modality of each attribute, compute the probability of belonging to each feature interval.

        .. math::
            P(Modality \\mid f \\in F_{i}) = P(f \\in F_{i} \\mid Modality) \\cdot \\frac{P(Modality)}{P(f \\in F_{i})}
        """

        # compute constraints on each modality
        attributes = functions.get_attributes(self.modalities)

        ech = {}
        constraints = {}
        for attribute in attributes:
            ech[attribute] = {}
            # get attribute frequency
            attribute_freq = self.crossed_modalities_frequencies.groupby(
                [attribute], as_index=False
            )["probability"].sum()
            for modality in self.modalities[attribute]:
                self.log("Computing constraints for modality: {} = {}".format(attribute, modality))

                # compute probability of each feature interval when being in a modality
                ech[attribute][modality] = self._compute_feature_prob(attribute, modality)

                # multiply frequencies by each element of ech_compo
                value = attribute_freq[attribute_freq[attribute].isin([modality])]
                if len(value) > 0:
                    probability = value["probability"].iloc[0]
                else:
                    probability = 0
                df = ech[attribute][modality]
                # prob(feature | modality) * frequency // ech is modified inplace here
                df["prob"] = df["prob"] * float(probability)

            ech_list = []
            for modality in ech[attribute]:
                ech_list.append(ech[attribute][modality])
            C = pd.concat(
                ech_list,
                axis=1,
            )
            # Somme P(feature & modality) sur les modality = P(feature)
            C = C.iloc[:, 1::2]
            C.columns = list(range(0, len(ech[attribute])))
            C["Proba"] = C.sum(axis=1)
            p = C[["Proba"]]

            # constraint
            constraints[attribute] = {}
            for modality in ech[attribute]:
                constraints[attribute][modality] = ech[attribute][modality]["prob"] / p["Proba"]

        self.constraints = constraints

    def _compute_feature_prob(self, attribute, modality):
        """
        Compute the probability of being in each feature interval with the given modality.

        :param attribute: attribute name
        :param modality: attribute modality

        :return: DataFrame
        """

        decile_tmp = self.distributions[
            self.distributions["modality"].isin([modality])
            & self.distributions["attribute"].isin([attribute])
        ]

        total_population_decile_tmp = [
            float(decile_tmp["D1"].iloc[0]),
            float(decile_tmp["D2"].iloc[0]),
            float(decile_tmp["D3"].iloc[0]),
            float(decile_tmp["D4"].iloc[0]),
            float(decile_tmp["D5"].iloc[0]),
            float(decile_tmp["D6"].iloc[0]),
            float(decile_tmp["D7"].iloc[0]),
            float(decile_tmp["D8"].iloc[0]),
            float(decile_tmp["D9"].iloc[0]),
            self.feature_values[-1],
        ]

        prob_df = functions.compute_features_prob(self.feature_values, total_population_decile_tmp)

        return prob_df

    def _get_feature_probs(self):
        """
        For each crossed modality, compute the probability of belonging to a feature interval.

        Invert the crossed modality probabilities using Bayes.

        Compute

        .. math::

            P(f \\in F_{i} \\mid M_{k}) = P(M_{k} \\mid f \\in F_{i}) \\cdot \\frac{P(f \\in F_{i})}{P(M_{k})}

        :return: DataFrame
        """

        feature_probs = self._compute_feature_probabilities_from_distributions()

        res = self.optim_result

        nb_columns = len(res.columns)

        for c in res.columns:
            res[c] = res[c].map(lambda x: x[0])  # POV sur conseils de Valentin

        for i in range(nb_columns):
            res[i] = res[i] * feature_probs["prob"][i]

        for i in range(len(res)):
            total = res.iloc[i, :].sum()
            res.iloc[i, :] = res.iloc[i, :] / total
        pd.set_option("display.max_rows", 500)
        pd.set_option("display.max_columns", 500)
        pd.set_option("display.width", 1000)

        # cumulated_results = res.to_numpy()

        # matrix = []
        #
        # for l in range(len(res)):
        #     last = cumulated_results[l, 0]
        #     row = [last]
        #     for i in range(1, self.nb_features):
        #         prob = cumulated_results[l, i] - last
        #         if prob <= 0:
        #             row.append(np.nan)
        #         else:
        #             row.append(prob)
        #
        #             last = cumulated_results[l, i]
        #
        #     matrix.append(row)
        #
        # probs = pd.DataFrame(matrix, columns=res.columns)

        return res

    def assign_feature_value_to_pop(self):
        """
        Assign feature values to the population individuals using the algorithm results.


        :return: enriched population DataFrame
        """

        self.log("Drawing feature values for the population", lg.INFO)

        # compute the probability of being in each feature interval, for each crossed modality
        res = self._get_feature_probs()

        # associate each individual to a crossed modality
        self.crossed_modalities_frequencies["index"] = self.crossed_modalities_frequencies.index
        merge = self.population.merge(
            self.crossed_modalities_frequencies,
            how="left",
            on=functions.get_attributes(self.modalities),
        )

        # associate a feature value to the population individuals
        merge["feature"] = merge["index"].apply(lambda x: self._draw_feature(res, x))

        # remove irrelevant columns
        merge.drop(["index", "probability"], axis=1, inplace=True)

        return merge

    def _draw_feature(self, res, index):
        """
        Draw a feature value using the given distribution.

        :param res: feature distributions by crossed modality
        :param index: index of the crossed modality

        :return: Drawn feature value
        """

        # get probs
        probs = res.loc[index,].to_numpy()
        interval_values = [self.parameters["abs_minimum"]] + self.feature_values

        # get the non-null probs and values
        probs2 = []
        values2 = [self.parameters["abs_minimum"]]
        for i in range(len(probs)):
            if pd.isna(probs[i]):
                continue
            probs2.append(probs[i])
            values2.append(interval_values[i + 1])
        # TODO : probs don't sum to 1, this isn't reassuring

        # draw a feature interval using the probs
        values = list(range(len(probs2)))
        feature_interval = random.choices(values, probs2)[0]
        assert len(probs2) == len(values2) - 1

        # draw a feature value using a uniform distribution in the interval
        lower, upper = values2[feature_interval], values2[feature_interval + 1]

        draw = random.random()

        final = round(
            lower + (upper - lower) * draw
        )  # POV : pourquoi on arrondit ? Cela n'explique pas les problèmes

        return final

    def _compute_feature_probabilities_from_distributions(self):
        """
        Compute the probability of each feature interval.

        Use the global distribution of the features to interpolate the interval probabilities.

        The resulting DataFrame contains :math:`P(f \\in F_{i})` for i in N

        :return: DataFrame
        """
        distrib_all_df = self.distributions[self.distributions["attribute"] == "all"]

        assert len(distrib_all_df) == 1

        total_population_decile = [
            self.distributions["D1"].iloc[0],
            self.distributions["D2"].iloc[0],
            self.distributions["D3"].iloc[0],
            self.distributions["D4"].iloc[0],
            self.distributions["D5"].iloc[0],
            self.distributions["D6"].iloc[0],
            self.distributions["D7"].iloc[0],
            self.distributions["D8"].iloc[0],
            self.distributions["D9"].iloc[0],
            self.feature_values[-1],  # maximum value of all deciles
        ]

        prob_df = functions.compute_features_prob(self.feature_values, total_population_decile)

        return prob_df

    # utils methods

    def log(message: str, level: int = lg.DEBUG):
        """
        Log a message using the package logger.

        See logging library.

        :param message: message to be logged
        :param level: logging level
        """

        utils.log(message, level)

    log = staticmethod(log)



"""
This is the main bhepop2 package.
"""

#: bhepop2 version
__version__ = "0.2.1"

# from bhepop2.bhepop2_enrichment import QuantitativeEnrichment
#
#
# # enrichment functions (avoids direct use of classes)
#
# def quantitative_enrichment(population, distributions, attribute_selection=None, parameters=None, seed=None):
#     """
#     Enrich the given synthetic population with a quantitative feature using marginal distributions.
#
#     :param population:
#     :param distributions:
#     :param attribute_selection:
#     :param parameters:
#     :param seed:
#
#     :return: enriched population
#     """
#     # create enrichment class instance
#     enrich_class = QuantitativeEnrichment(
#         population,
#         distributions,
#         attribute_selection=attribute_selection,
#         parameters=parameters,
#         seed=seed,
#     )
#
#     # run entropy maximisation
#     enrich_class.optimise()
#
#     # return enriched population
#     return enrich_class.assign_feature_value_to_pop()
#
